const state = require('../../state');
const {keyboardMap, modifierMap, lockMap, location} = require('./mapping');
const {createKeyboardEvent} = require('../../event');
const page = require('../../page');
const {Button} = require('../device');

const REPEAT_TIMEOUT = 500;
const REPEAT_INTERVAL = 500;

state.register('keyboard-lock', {
	CapsLock: false,
	NumLock: false,
	ScrollLock: false
});

class KeyboardButton extends Button {
	constructor(options, {code, key, keycode, location, shift = key, press = null}) {
		super(options);

		this.code = code;
		this.key = key;
		this.keycode = keycode;
		this.location = location;
		this.press = press;
		this.shift = shift;

		this.$activeRepeatTimeout = null;
		this.$repeatInterval = null;
	}

	down() {
		return super.down().then(() => {
			// 小键盘锁定
			if(state.get('keyboard-lock').NumLock && this.location === location.DOM_KEY_LOCATION_NUMPAD) {
				return;
			}
			this.$dispatch('keydown');
			if (this.press) {
				this.$dispatch('keypress');
			}
		});
	}

	up() {
		return this.$checkReleased().then(() => {
			if(!(state.get('keyboard-lock').NumLock && this.location === location.DOM_KEY_LOCATION_NUMPAD)) {
				this.$dispatch('keyup');
			}
			this.upFn();
		}, this.$stdOut);
	}

	$dispatch(eventType) {
		let params = {
			args: {
				key: state.get('modifier').shift^state.get('keyboard-lock').CapsLock ? this.shift : this.key,
				code: this.code,
				location: this.location
			},
			readOnly: {
				keyCode: this.keycode,
				which: this.keycode,
				charCode: 0
			}
		};
		if(eventType === 'keypress') {
			const presscode = state.get('modifier').shift^state.get('keyboard-lock').CapsLock ? this.press.shift : this.press.keycode;
			params.readOnly = {
				keyCode: presscode,
				which: presscode,
				charCode: presscode,
			};
		}
		page.dispatchEvent(createKeyboardEvent(eventType, params));
	}
}

let keyboard = {};
for(let code in keyboardMap) {
	const keyButton = keyboardMap[code];
	keyButton.code = code;
	keyboard[code] = new KeyboardButton({
		down() {
			this.$activeRepeatTimeout = setTimeout(() => {
				this.$repeatInterval = setInterval(() => {
					keyboard[code].down();
				}, REPEAT_INTERVAL);
			}, REPEAT_TIMEOUT);
		},
		up() {
			clearTimeout(this.$activeRepeatTimeout);
			clearInterval(this.$repeatInterval);
		}
	}, keyButton);
}

for(let code in modifierMap) {
	const keyButton = modifierMap[code];
	keyButton.code = code;
	keyboard[code] = new KeyboardButton({
		down() {
			state.update('modifier', clone => {
				clone[keyButton.modifier] = true;
			});
		},
		up() {
			state.update('modifier', clone => {
				clone[keyButton.modifier] = false;
			});
		}
	}, keyButton);
}

for(let code in lockMap) {
	const keyButton = lockMap[code];
	keyButton.code = code;
	keyboard[code] = new KeyboardButton({
		down() {
			state.update('modifier', clone => {
				clone[code] = !clone[code];
			});
		}
	}, keyButton);
}

module.exports = {
	keyboard: keyboard,
	down(keyName) {
		return this.keyboard[keyName].down();
	},
	up(keyName) {
		return this.keyboard[keyName].up();
	},
	press(keyName) {
		return Promise.resolve().then(() => {
			return this.keyboard[keyName].down(keyName);
		}).then(() => {
			return this.keyboard[keyName].up(keyName);
		});
	}
};