const {createUIEvent} = require('../event');
const {keyboard} = require('../device');

function isInputElement(element) {
	return element.tagName === 'INPUT';
}

function isTextAreaElement(element) {
	return element.tagName === 'TEXTAREA';
}

function isContentEditable(element) {
	return element.isContentEditable;
}

/**
 * @param {any} element Maybe it is a html element.
 */
function isSelectElement(element) {
	return element.tagName === 'SELECT';
}

const textStyleInputElement = [
	// Typical text style.
	'text', 'password', 'email', 'search', 'tel', 'url',
	// Widget style
	'number', 'color', 'range',
	// date
	'date', 'month', 'week', 'time', 'datetime', 'datetime-local'
];
function isTextStyle(element) {
	if (isInputElement(element)) {
		return Boolean(~textStyleInputElement.indexOf(element.type));
	}
	return false;
}

const switchStyleInputElement = ['radio', 'checkbox'];
function isSwitchStyle(element) {
	if (isInputElement(element)) {
		return Boolean(~switchStyleInputElement.indexOf(element.type));
	}
	return false;
}

/**
 * noEditableInputElement
 *  - ['button', 'submit', 'hidden', 'reset', 'image'];
 */
function isEditable(element) {
	return element.disabled === false && element.readOnly === false;
}

class Accessor {
	constructor (element) {
		this.element = element;
	}

	setValue() {
		return null;
	}

	getValue() {
		return this.element.innerText;
	}

	getStyle(cssProperty) {

	}

	getAttr(attrName) {

	}

	get isDisplay() {
		const {offsetWidth, offsetHeight} = this.element;
		return offsetWidth * offsetHeight > 0;
	}
}

class TextAccessor extends Accessor {
	setValue(value) {
		if (!isEditable(this.element)) {
			return false;
		}

		this.element.value = String(value);

		// TODO Maybe used in Firefox
		// if (window.InputEvent) {
			
		// }
		this.element.dispatchEvent(createUIEvent('input'));
		keyboard.pressKey('ControlLeft');
		this.element.dispatchEvent(createUIEvent('change'));

		return true;
	}

	getValue() {
		return this.element.value;
	}
}

class SwitchAccessor extends Accessor {
	setValue(value) {
		if (!isEditable(this.element)) {
			return false;
		}

		const oldValue = this.element.checked;

		if (oldValue !== Boolean(value)) {
			this.element.checked = Boolean(value);
			this.element.dispatchEvent(createUIEvent('change'));
		}

		return true;
	}

	getValue() {
		return this.element.checked;
	}
}

class SelectAccessor extends Accessor {
	$getOptionElement(value) {
		var head = 0,
			len = this.element.length,
			HEADOptionHTML = null;

		if (typeof value === 'number') {
			return this.element[value];
		}
		
		value += '';
		for (; head < len; head += 1) {
			HEADOptionHTML = this.element[head];
			if (HEADOptionHTML.value === value) {
				break;
			}
		}

		return HEADOptionHTML;
	}

	setValue(value) {
		if (this.element.disabled) {
			return false;
		}
		
		const optionElement = this.$getOptionElement(value);

		if (!optionElement) {
			return false;
		}

		optionElement.selected = true;

		this.element.dispatchEvent(createUIEvent('change'));
		
		return true;
	}

	getValue() {
		return this.element.value;
	}
}

class ContentAccessor extends Accessor {
	setValue(value) {

		this.element.textContent += String(value);
		
		if (window.InputEvent) {
			this.element.dispatchEvent(createUIEvent('input'));
		}

		return true;
	}

	getValue() {
		return this.element.textContent;
	}
}

module.exports = function (element) {
	if (isTextAreaElement(element) || isTextStyle(element)) {
		return new TextAccessor(element);
	}

	if (isSwitchStyle(element)) {
		return new SwitchAccessor(element);
	}

	if (isSelectElement(element)) {
		return new SelectAccessor(element);
	}

	if (isContentEditable(element)) {
		return new ContentAccessor(element);
	}

	return new Accessor(element);
};