/**
 * Get the top, left, width, height of element.
 */
function getElementFrameElement(element) {
	return element.ownerDocument.defaultView.frameElement;
}
function cloneRect({bottom, height, left, right, top, width}) {
	return {bottom, height, left, right, top, width};
}
function getElementRect(element, crossFrame = true) {
	const rect = cloneRect(element.getBoundingClientRect());
	while (crossFrame && (element = getElementFrameElement(element))) {
		const frameRect = element.getBoundingClientRect();
		rect.top += frameRect.top;
		rect.left += frameRect.left;
	}
	return rect;
}

const MODE = {
	CENTER: 0,
	RANDOM: 1
};

function getElementClientPosition(element, mode = MODE.CENTER) {
	const rect = getElementRect(element);

	const maxHeight = rect.height;
	const maxWidth = rect.width;

	let offset;
	if (mode === MODE.CENTER) {
		offset = {
			top: Math.floor(maxHeight / 2),
			left: Math.floor(maxWidth / 2)
		};
	} else if (mode === MODE.RANDOM) {
		const seed = Math.random();
		offset = {
			top: Math.floor(seed * maxHeight),
			left: Math.floor(seed * maxWidth),
		};
	}
	const client = {
		top: offset.top + rect.top,
		left: offset.left + rect.left
	};

	return {
		clientX: client.left,
		clientY: client.top
	};
}

/**
 * Try to make the element into the vision.
 * 
 * @param {Element} element The element which to be watched.
 */
function scrollByElement(element) {
	return Promise.resolve().then(() => {
		element.scrollIntoViewIfNeeded && element.scrollIntoViewIfNeeded(true);
		
		return element;
	});
}

function elementFromPoint(left, top, isCrossIframe = true) {
	//TODO check cross-domain
	const iframeStack = [];

	let element = document.elementFromPoint(left, top);
	while (isCrossIframe && element && element.tagName.match('FRAME')) {
		let doc, iframe, rect;

		iframeStack.push(iframe = element);
		rect = iframe.getBoundingClientRect();
		doc = iframe.contentDocument;
		element = doc.elementFromPoint(left -= rect.left, top -= rect.top);
	}

	return element || document;
}

function elementAccessble(element) {
	const rect = getElementRect(element);
	if(rect.width*rect.height === 0) return false;
	const position = getElementClientPosition(element);
	const target = elementFromPoint(position.clientX, position.clientY);
	if(element !== target) return false;
	return true;
}

function queryCommonParent(element_0, element_1) {
	if (!element_0 || !element_1) {
		return document;
	}

	const parentList_0 = queryParentPath(element_0);
	const parentList_1 = queryParentPath(element_1);
	
	let parent = parentList_1.shift();
	while (parent && parentList_0.indexOf(parent) === -1) {
		parent = parentList_1.shift();
	}

	return parent;
}

function queryParentPath(element, parent = null) {
	if (!(element && 
		(element.nodeType === 1 || element.nodeType === 9))) {
			
		throw new Error('Invalid element.');
	}

	const parentPath = [];
	while (element && element !== parent &&
		element.parentElement !== element.ownerDocument) {
		parentPath.push(element);
		element = element.parentElement;
	}

	return parentPath;
}

const canFocusedElement =
	['A', 'AREA', 'LABEL', 'INPUT', 'SELECT', 'TEXTAREA', 'BUTTON'];
function focusable(element) {
	return element && ~canFocusedElement.indexOf(element.tagName);
}

exports.elementAccessble = elementAccessble;
exports.fromPoint = elementFromPoint;
exports.queryCommonParent = queryCommonParent;
exports.queryParent = queryParentPath;
exports.getElementClientPosition = getElementClientPosition;
exports.getElementRect = getElementRect;
exports.scrollByElement = scrollByElement;
exports.focusable = focusable;